import pandas as pd
import geopandas as gpd
from ..utils import constants, utils, plot
import numpy as np
from warnings import warn
from ..tessellation.tilers import tiler
from shapely.geometry import Point, Polygon


class FlowSeries(pd.Series):

    @property
    def _constructor(self):
        return FlowSeries

    @property
    def _constructor_expanddim(self):
        return FlowDataFrame


class FlowDataFrame(pd.DataFrame):
    """
    A FlowDataFrame object is a pandas.DataFrame that has three columns origin, destination, and flow. FlowDataFrame accepts the following keyword arguments:
    
    Parameters
    ----------
    data : list or dict or pandas DataFrame
        the data that must be embedded into a FlowDataFrame.
        
    origin : str, optional
        the name of the column in `data` containing the origin location. The default is `constants.ORIGIN`.
        
    destination : str, optional
        the name of the column in `data` containing the destination location. The default is `constants.DESTINATION`.
        
    flow : str, optional
        the name of the column in `data` containing the flow between two locations. The default is `constants.FLOW`.
        
    datetime : str, optional
        the name of the column in `data` containing the datetime the flow refers to. The default is `constants.DATETIME`.
        
    tile_id : std, optional
        the name of the column in `data` containing the tile identifier. The default is `constants.TILE_ID`.
        
    timestamp : boolean, optional
        it True, the datetime is a timestamp. The default is `False`.
        
    tessellation : GeoDataFrame, optional
        the spatial tessellation on which the flows take place. The default is `None`.
        
    parameters : dict, optional
        parameters to add to the FlowDataFrame. The default is `{}` (no parameters).
        
    Examples
    --------
    >>> import skmob
    >>> import geopandas as gpd
    >>> # load a spatial tessellation
    >>> url_tess = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
    >>> tessellation = gpd.read_file(url_tess).rename(columns={'tile_id': 'tile_ID'})
    >>> print(tessellation.head())
      tile_ID  population                                           geometry
    0   36019       81716  POLYGON ((-74.006668 44.886017, -74.027389 44....
    1   36101       99145  POLYGON ((-77.099754 42.274215, -77.0996569999...
    2   36107       50872  POLYGON ((-76.25014899999999 42.296676, -76.24...
    3   36059     1346176  POLYGON ((-73.707662 40.727831, -73.700272 40....
    4   36011       79693  POLYGON ((-76.279067 42.785866, -76.2753479999...    
    >>> # load real flows into a FlowDataFrame
    >>> # download the file with the real fluxes from: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_commuting_flows_2011.csv
    >>> fdf = skmob.FlowDataFrame.from_file("NY_commuting_flows_2011.csv", 
                                            tessellation=tessellation, 
                                            tile_id='tile_ID', 
                                            sep=",")
    >>> print(fdf.head())
         flow origin destination
    0  121606  36001       36001
    1       5  36001       36005
    2      29  36001       36007
    3      11  36001       36017
    4      30  36001       36019      
    """
    _metadata = ['_tessellation', '_parameters']

    def __init__(self, data, origin=constants.ORIGIN, destination=constants.DESTINATION, flow=constants.FLOW,
                 datetime=constants.DATETIME, tile_id=constants.TILE_ID, timestamp=False, tessellation=None,
                 parameters={}):

        original2default = {origin: constants.ORIGIN,
                            destination: constants.DESTINATION,
                            flow: constants.FLOW,
                            datetime: constants.DATETIME}

        columns = None

        if isinstance(data, pd.DataFrame):
            fdf = data.rename(columns=original2default)
            columns = fdf.columns

        # Dictionary
        elif isinstance(data, dict):
            fdf = pd.DataFrame.from_dict(data).rename(columns=original2default)
            columns = fdf.columns

        # List
        elif isinstance(data, list) or isinstance(data, np.ndarray):
            fdf = data
            columns = []
            num_columns = len(data[0])
            for i in range(num_columns):
                try:
                    columns += [original2default[i]]
                except KeyError:
                    columns += [i]

        elif isinstance(data, pd.core.internals.BlockManager):
            fdf = data

        else:
            raise TypeError('DataFrame constructor called with incompatible data and dtype: {e}'.format(e=type(data)))

        super(FlowDataFrame, self).__init__(fdf, columns=columns)

        if parameters is None:
            # Init empty prop dictionary
            self._parameters = {}
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise AttributeError("Parameters must be a dictionary.")

        if not isinstance(data, pd.core.internals.BlockManager):

            self[constants.ORIGIN] = self[constants.ORIGIN].astype('str')
            self[constants.DESTINATION] = self[constants.DESTINATION].astype('str')

            if tessellation is None:
                raise TypeError("tessellation must be a GeoDataFrame with tile_id and geometry.")

            elif isinstance(tessellation, gpd.GeoDataFrame):
                self._tessellation = tessellation.copy()
                self._tessellation.rename(columns={tile_id: constants.TILE_ID}, inplace=True)
                self._tessellation[constants.TILE_ID] = self._tessellation[constants.TILE_ID].astype('str')

                if tessellation.crs is None:
                    warn("The tessellation crs is None. It will be set to the default crs WGS84 (EPSG:4326).")

                # Check consistency
                origin = self[constants.ORIGIN]
                destination = self[constants.DESTINATION]

                if not all(origin.isin(self._tessellation[constants.TILE_ID])) or \
                        not all(destination.isin(self._tessellation[constants.TILE_ID])):
                    raise ValueError("Inconsistency - origin and destination IDs must be present in the tessellation.")

                # Cleaning the index to make sure it is incremental
                self._tessellation.reset_index(inplace=True, drop=True)

            else:
                raise TypeError("tessellation must be a GeoDataFrame with tile_id and geometry.")

            if self._has_flow_columns():
                self._set_flow(timestamp=timestamp, inplace=True)

    def get_flow(self, origin_id, destination_id):
        """
        Get the flow between two locations. If there is no flow between two locations it returns 0.
        
        Parameters
        ----------
        origin_id : str 
            the identifier of the origin tile.
            
        destination_id : str
            the identifier of the tessellation tile.
            
        Returns
        -------
        int
            the flow between the two locations.
        
        Examples
        --------
        >>> import skmob
        >>> import geopandas as gpd
        >>> # load a spatial tessellation
        >>> url_tess = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
        >>> tessellation = gpd.read_file(url_tess).rename(columns={'tile_id': 'tile_ID'})
        >>> print(tessellation.head())
          tile_ID  population                                           geometry
        0   36019       81716  POLYGON ((-74.006668 44.886017, -74.027389 44....
        1   36101       99145  POLYGON ((-77.099754 42.274215, -77.0996569999...
        2   36107       50872  POLYGON ((-76.25014899999999 42.296676, -76.24...
        3   36059     1346176  POLYGON ((-73.707662 40.727831, -73.700272 40....
        4   36011       79693  POLYGON ((-76.279067 42.785866, -76.2753479999...    
        >>> # load real flows into a FlowDataFrame
        >>> # download the file with the real fluxes from: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_commuting_flows_2011.csv
        >>> fdf = skmob.FlowDataFrame.from_file("NY_commuting_flows_2011.csv", 
                                                tessellation=tessellation, 
                                                tile_id='tile_ID', 
                                                sep=",")
        >>> print(fdf.head())
             flow origin destination
        0  121606  36001       36001
        1       5  36001       36005
        2      29  36001       36007
        3      11  36001       36017
        4      30  36001       36019 
        >>> flow = fdf.get_flow('36001', '36007')
        >>> print(flow)
        29
        """

        if (origin_id not in self._tessellation[constants.TILE_ID].values) or \
                (destination_id not in self._tessellation[constants.TILE_ID].values):
            raise ValueError("Both origin_id and destination_id must be present in the tessellation.")

        tmp = self[(self[constants.ORIGIN] == origin_id) & (self[constants.DESTINATION] == destination_id)]
        if len(tmp) == 0:
            return 0
        else:
            return tmp[constants.FLOW].iloc[0]

    def settings_from(self, flowdataframe):
        """
        Copy the attributes from another FlowDataFrame.
        
        Parameters
        ----------
        flowdataframe : FlowDataFrame 
            the FlowDataFrame from which to copy the attributes.
        """
        for k in flowdataframe.metadata:
            value = getattr(flowdataframe, k)
            setattr(self, k, value)

    def get_geometry(self, tile_id):
        if tile_id not in self._tessellation[constants.TILE_ID].values:
            raise ValueError("tile_id \"%s\" is not in the tessellation." % tile_id)

        # selecting from geopandas will return a series with one element, for this reason we use iloc to get the object
        return self.tessellation[self.tessellation[constants.TILE_ID] == tile_id].geometry.iloc[0]

    def to_matrix(self):

        m = np.zeros((len(self._tessellation), len(self._tessellation)))

        def _to_matrix(df, matrix, tessellation):
            o = tessellation.index[tessellation['tile_ID'] == df['origin']].iloc[0]
            d = tessellation.index[tessellation['tile_ID'] == df['destination']].iloc[0]

            matrix[o][d] = df['flow']

        self.apply(_to_matrix, args=(m, self._tessellation), axis=1)

        return m

    def _has_flow_columns(self):

        if (constants.ORIGIN in self) and (constants.DESTINATION in self) and (constants.FLOW in self):
            return True

        return False

    def _is_flowdataframe(self):

        if ((constants.ORIGIN in self) and
                pd.core.dtypes.common.is_string_dtype(self[constants.ORIGIN])) \
            and ((constants.DESTINATION in self) and
                 pd.core.dtypes.common.is_string_dtype(self[constants.DESTINATION])) \
            and ((constants.TILE_ID in self._tessellation) and
                 pd.core.dtypes.common.is_string_dtype(self._tessellation[constants.TILE_ID])) \
            and ((constants.FLOW in self) and
                 (pd.core.dtypes.common.is_float_dtype(self[constants.FLOW]) or
                  pd.core.dtypes.common.is_integer_dtype(self[constants.FLOW]))):
            return True

        return False


    def _set_flow(self, timestamp=False, inplace=False):

        if not inplace:
            frame = self.copy()
        else:
            frame = self

        if timestamp:
            frame[constants.DATETIME] = pd.to_datetime(frame[constants.DATETIME], unit='s')

        frame.parameters = self._parameters
        frame.tessellation = self._tessellation

        # Set dtypes on columns
        if not pd.core.dtypes.common.is_string_dtype(frame._tessellation[constants.TILE_ID]):
            frame._tessellation[constants.TILE_ID] = frame._tessellation[constants.TILE_ID].astype('str')

        if not pd.core.dtypes.common.is_string_dtype(frame[constants.ORIGIN]):
            frame._tessellation[constants.ORIGIN] = frame._tessellation[constants.ORIGIN].astype('str')

        if not pd.core.dtypes.common.is_string_dtype(frame[constants.DESTINATION]):
            frame._tessellation[constants.DESTINATION] = frame._tessellation[constants.DESTINATION].astype('str')

        if not inplace:
            return frame

    def __getitem__(self, key):
        """
        It the result contains lat, lng and datetime, return a TrajDataFrame, else a pandas DataFrame.
        """
        result = super(FlowDataFrame, self).__getitem__(key)

        if (isinstance(result, FlowDataFrame)) and result._is_flowdataframe():
            result.__class__ = FlowDataFrame
            result.tessellation = self._tessellation
            result.parameters = self._parameters

        elif isinstance(result, FlowDataFrame) and not result._is_flowdataframe():
            result.__class__ = pd.DataFrame

        return result

    @classmethod
    def from_file(cls, filename, origin=None, destination=None, origin_lat=None, origin_lng=None, destination_lat=None,
                  destination_lng=None, flow=constants.FLOW, datetime=constants.DATETIME, timestamp=False, sep=",",
                  tessellation=None, tile_id=constants.TILE_ID, usecols=None, header='infer', parameters=None,
                  remove_na=False):

        # Case 1: origin, destination, flow, [datetime]
        if (origin is not None) and (destination is not None):

            if not isinstance(tessellation, gpd.GeoDataFrame):
                raise AttributeError("tessellation must be a GeoDataFrame.")

        df = pd.read_csv(filename, sep=sep, header=header, usecols=usecols)

        # Case 2: origin_lat, origin_lng, destination_lat, destination_lng, flow, [datetime]
        if (origin_lat is not None) and (origin_lng is not None) and (destination_lat is not None) and \
                (destination_lng is not None):

            # Step 1: if tessellation is None infer it from data
            if tessellation is None:

                a = df[[origin_lat, origin_lng]].rename(columns={origin_lat: 'lat', origin_lng: 'lng'})

                b = df[[destination_lat, destination_lng]].rename(columns={destination_lat: 'lat',
                                                                           destination_lng: 'lng'})

                # DropDuplicates has to be applied now because Geopandas doesn't support removing duplicates in geometry
                points = pd.concat([a, b]).drop_duplicates(['lat', 'lng'])
                points = gpd.GeoDataFrame(geometry=gpd.points_from_xy(points['lng'], points['lat']),
                                          crs=constants.DEFAULT_CRS)

                tessellation = tiler.get('voronoi', points=points)

            # Step 2: map origin and destination points into the tessellation

            gdf_origin = gpd.GeoDataFrame(df.copy(), geometry=gpd.points_from_xy(df[origin_lng], df[origin_lat]),
                                          crs=tessellation.crs)
            gdf_destination = gpd.GeoDataFrame(df.copy(),
                                               geometry=gpd.points_from_xy(df[destination_lng], df[destination_lat]),
                                               crs=tessellation.crs)

            if all(isinstance(x, Polygon) for x in tessellation.geometry):

                if remove_na:
                    how = 'inner'
                else:
                    how = 'left'

                origin_join = gpd.sjoin(gdf_origin, tessellation, how=how, op='within').drop("geometry", axis=1)
                destination_join = gpd.sjoin(gdf_destination, tessellation, how=how, op='within').drop("geometry",
                                                                                                       axis=1)

                df = df.merge(origin_join[[constants.TILE_ID]], left_index=True, right_index=True)
                df.loc[:, constants.ORIGIN] = origin_join[constants.TILE_ID]
                df.drop([constants.ORIGIN_LAT, constants.ORIGIN_LNG, constants.TILE_ID], axis=1, inplace=True)

                df = df.merge(destination_join[[constants.TILE_ID]], left_index=True, right_index=True)
                df.loc[:, constants.DESTINATION] = destination_join[constants.TILE_ID]
                df.drop([constants.DESTINATION_LAT, constants.DESTINATION_LNG, constants.TILE_ID], axis=1, inplace=True)

            elif all(isinstance(x, Point) for x in tessellation.geometry):

                df.loc[:, constants.ORIGIN] = utils.nearest(gdf_origin, tessellation, constants.TILE_ID).values
                df.loc[:, constants.DESTINATION] = utils.nearest(gdf_destination, tessellation,
                                                                 constants.TILE_ID).values

                df.drop([origin_lat, origin_lng, destination_lat, destination_lng], inplace=True, axis=1)

            else:
                raise AttributeError("In case of expanded format (coordinates instead of ids), the tessellation must "
                                     "contains either all Polygon or all Point. Mixed types are not allowed.")

        # Step 3: call the constructor

        if parameters is None:
            parameters = {'from_file': filename}

        return cls(df, origin=constants.ORIGIN, destination=constants.DESTINATION, flow=flow, datetime=datetime,
                   timestamp=timestamp, tessellation=tessellation, parameters=parameters, tile_id=tile_id)

    @property
    def origin(self):
        if constants.ORIGIN not in self:
            raise AttributeError("The FlowDataFrame does not contain the column '%s.'" % constants.ORIGIN)

        return self[constants.ORIGIN]

    @property
    def destination(self):
        if constants.DESTINATION not in self:
            raise AttributeError("The FlowDataFrame does not contain the column '%s.'" % constants.DESTINATION)

        return self[constants.DESTINATION]

    @property
    def flow(self):
        if constants.FLOW not in self:
            raise AttributeError("The FlowDataFrame does not contain the column '%s.'" % constants.FLOW)

        return self[constants.FLOW]

    @property
    def datetime(self):
        if constants.DATETIME not in self:
            raise AttributeError("The FlowDataFrame does not contain the column '%s.'" % constants.DATETIME)

        return self[constants.DATETIME]

    @property
    def tessellation(self):
        return self._tessellation

    @tessellation.setter
    def tessellation(self, tessellation):
        self._tessellation = tessellation

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):

        self._parameters = dict(parameters)

    @property
    def metadata(self):

        md = ['crs', 'parameters', 'tessellation']    # Add here all the metadata that are accessible from the object
        return md

    @property
    def _constructor(self):
        return FlowDataFrame

    @property
    def _constructor_sliced(self):
        return FlowSeries

    @property
    def _constructor_expanddim(self):
        return FlowDataFrame

    # Plot methods
    def plot_flows(self, map_f=None, min_flow=0, tiles='Stamen Toner', zoom=6, flow_color='red', opacity=0.5,
                   flow_weight=5, flow_exp=0.5, style_function=plot.flow_style_function,
                   flow_popup=False, num_od_popup=5, tile_popup=True, radius_origin_point=5,
                   color_origin_point='#3186cc'):
        """
        Plot the flows of a FlowDataFrame on a Folium map.
        
        Parameters
        ----------
        map_f : folium.Map, optional
            the `folium.Map` object where the flows will be plotted. If `None`, a new map will be created. The default is `None`.

        min_flow : float, optional
            only flows larger than `min_flow` will be plotted. The default is `0`.

        tiles: str, optional
            folium's `tiles` parameter. The default is `Stamen Toner`.

        zoom : int, optional
            initial zoom of the map. The default is `6`. 

        flow_color : str, optional
            the color of the flow edges. The default is `red`.

        opacity : float, optional
            the opacity (alpha level) of the flow edges. The default is `0.5`.

        flow_weight : float, optional
            the weight factor used in the function to compute the thickness of the flow edges. The default is `5`.

        flow_exp : float, optional
            the weight exponent used in the function to compute the thickness of the flow edges. The default is `0.5`.

        style_function : lambda function, optional
            the GeoJson style function. The default is `plot.flow_style_function`.

        flow_popup : boolean, optional
            if `True`, when clicking on a flow edge a popup window displaying information on the flow will appear. The default is `False`.

        num_od_popup : int, optional
            number of origin-destination pairs to show in the popup window of each origin location. The default is `5`.

        tile_popup : boolean, optional
            if `True`, when clicking on a location marker a popup window displaying information on the flows departing from that location will appear. The default is `True`.

        radius_origin_point : float, optional
            the size of the location markers. The default is `5`.

        color_origin_point : str, optional
            the color of the location markers. The default is '#3186cc'.
        
        Returns
        -------
        folium.Map
            the `folium.Map` object with the plotted flows.
        
        Examples
        --------
        >>> import skmob
        >>> import geopandas as gpd
        >>> # load a spatial tessellation
        >>> url_tess = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
        >>> tessellation = gpd.read_file(url_tess).rename(columns={'tile_id': 'tile_ID'})    
        >>> # load real flows into a FlowDataFrame
        >>> # download the file with the real fluxes from: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_commuting_flows_2011.csv
        >>> fdf = skmob.FlowDataFrame.from_file("NY_commuting_flows_2011.csv", 
                                                tessellation=tessellation, 
                                                tile_id='tile_ID', 
                                                sep=",")
        >>> print(fdf.head())
             flow origin destination
        0  121606  36001       36001
        1       5  36001       36005
        2      29  36001       36007
        3      11  36001       36017
        4      30  36001       36019  
        >>> m = fdf.plot_flows(flow_color='red')
        >>> m
        
        .. image:: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/examples/plot_flows_example.png
        """
        return plot.plot_flows(self, map_f=map_f, min_flow=min_flow,  tiles=tiles, zoom=zoom, flow_color=flow_color,
                               opacity=opacity, flow_weight=flow_weight, flow_exp=flow_exp,
                               style_function=style_function, flow_popup=flow_popup, num_od_popup=num_od_popup,
                               tile_popup=tile_popup, radius_origin_point=radius_origin_point,
                               color_origin_point=color_origin_point)

    def plot_tessellation(self, map_osm=None, maxitems=-1, style_func_args={}, popup_features=[constants.TILE_ID],
                          tiles='Stamen Toner', zoom=6, geom_col='geometry'):
        """
        Plot the spatial tessellation on a Folium map.
        
        Parameters
        ----------
        map_osm : folium.Map, optional
            the `folium.Map` object where the GeoDataFrame describing the spatial tessellation will be plotted. If `None`, a new map will be created. The default is `None`.

        maxitems : int, optional
            maximum number of tiles to plot. If `-1`, all tiles will be plotted. The default is `-1`.

        style_func_args : dict, optional
            a dictionary to pass the following style parameters (keys) to the GeoJson style function of the polygons: 'weight', 'color', 'opacity', 'fillColor', 'fillOpacity'. The default is `{}`.

        popup_features : list, optional
            when clicking on a tile polygon, a popup window displaying the information in the
            columns of `gdf` listed in `popup_features` will appear. The default is `[constants.TILE_ID]`.

        tiles : str, optional
            folium's `tiles` parameter. The default is 'Stamen Toner'.

        zoom : int, optional
            the initial zoom of the map. The default is `6`.

        geom_col : str, optional
             the name of the geometry column of the GeoDataFrame representing the spatial tessellation. The default is 'geometry'.
        
        Returns
        -------
        folium.Map
            the `folium.Map` object with the plotted GeoDataFrame.
            
        Examples
        --------
        >>> import skmob
        >>> import geopandas as gpd
        >>> # load a spatial tessellation
        >>> url_tess = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
        >>> tessellation = gpd.read_file(url_tess).rename(columns={'tile_id': 'tile_ID'})    
        >>> # load real flows into a FlowDataFrame
        >>> # download the file with the real fluxes from: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_commuting_flows_2011.csv
        >>> fdf = skmob.FlowDataFrame.from_file("NY_commuting_flows_2011.csv", 
                                                tessellation=tessellation, 
                                                tile_id='tile_ID', 
                                                sep=",")
        >>> m = fdf.plot_tessellation(popup_features=['tile_ID', 'population'])
        >>> m
        
        .. image:: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/examples/plot_tessellation_example.png
        """
        return plot.plot_gdf(self.tessellation, map_osm=map_osm, maxitems=maxitems, style_func_args=style_func_args,
                             popup_features=popup_features, tiles=tiles, zoom=zoom, geom_col=geom_col)



from abc import ABC, abstractmethod
import math
import geopandas as gpd
from shapely.geometry import Polygon, Point
from shapely.ops import cascaded_union
from ..utils import constants, utils
import numpy as np


class TessellationTilers:

    def __init__(self):

        self._tilers = {}

    def register_tiler(self, key, tiler):

        self._tilers[key] = tiler

    def create(self, key, **kwargs):

        tiler = self._tilers.get(key)

        if not tiler:
            raise ValueError(key)

        return tiler(**kwargs)

    def get(self, service_id, **kwargs):

        return self.create(service_id, **kwargs)


tiler = TessellationTilers()


class TessellationTiler(ABC):

    @abstractmethod
    def __call__(self, **kwargs):
        pass

    @abstractmethod
    def _build(self, **kwargs):
        pass


class VoronoiTessellationTiler(TessellationTiler):

    def __init__(self):

        super().__init__()
        self._instance = None

    def __call__(self, points, crs=constants.DEFAULT_CRS):

        if not self._instance:

            if isinstance(points, gpd.GeoDataFrame):

                if not all(isinstance(x, Point) for x in points.geometry):
                    raise ValueError("Not valid points object. Accepted type is GeoDataFrame.")

        return self._build(points, crs)

    def _build(self, points, crs=constants.DEFAULT_CRS):

        gdf = gpd.GeoDataFrame(points.copy(), crs=crs)
        gdf.loc[:, constants.TILE_ID] = list(np.arange(0, len(gdf)))

        # Convert TILE_ID to have str type
        gdf[constants.TILE_ID] = gdf[constants.TILE_ID].astype('str')

        return gdf[[constants.TILE_ID, 'geometry']]


# Register the builder
tiler.register_tiler('voronoi', VoronoiTessellationTiler())


class SquaredTessellationTiler(TessellationTiler):

    def __init__(self):

        super().__init__()
        self._instance = None

    def __call__(self, base_shape, meters=50, crs=constants.DEFAULT_CRS, window_size=None):
        if not self._instance:

            if isinstance(base_shape, str):
                # Try to obatain the base shape from OSM
                base_shape = utils.bbox_from_name(base_shape)

            elif isinstance(base_shape, gpd.GeoDataFrame) or isinstance(base_shape, gpd.GeoSeries):

                if all(isinstance(x, Point) for x in base_shape.geometry):
                    # Build a base shape that contains all the points in the given geodataframe
                    base_shape = utils.bbox_from_points(base_shape)

                elif all(isinstance(x, Polygon) for x in base_shape.geometry) and len(base_shape) > 1:

                    # Merge all the polygons
                    polygons = base_shape.geometry.values
                    base_shape = gpd.GeoSeries(cascaded_union(polygons), crs=base_shape.crs)

                #elif not all(isinstance(x, Polygon) for x in base_shape.geometry):
                #    raise ValueError("Not valid geometry object. Accepted types are Point and Polygon.")
            else:
                raise ValueError("Not valid base_shape object. Accepted types are str, GeoDataFrame or GeoSeries.")

        return self._build(base_shape, meters, crs)

    def _build(self, base_shape, meters, crs=constants.DEFAULT_CRS):

        # We work with the universal crs epsg:3857
        tmp_crs = constants.UNIVERSAL_CRS
        tmp_crs['units'] = 'm'

        area = base_shape.to_crs(tmp_crs)

        # Obtain the boundaries of the geometry
        boundaries = dict({'min_x': area.total_bounds[0],
                           'min_y': area.total_bounds[1],
                           'max_x': area.total_bounds[2],
                           'max_y': area.total_bounds[3]})

        # Find number of square for each side
        x_squares = int(math.ceil(math.fabs(boundaries['max_x'] - boundaries['min_x']) / meters))
        y_squares = int(math.ceil(math.fabs(boundaries['min_y'] - boundaries['max_y']) / meters))

        # Placeholder for the polygon
        polygons = []

        shape = area.unary_union

        # Iterate on the x
        for i in range(0, x_squares):

            # Increment x
            x1 = boundaries['min_x'] + (meters * i)
            x2 = boundaries['min_x'] + (meters * (i + 1))

            # Iterate on y
            for j in range(0, y_squares):

                # Increment y
                y1 = boundaries['min_y'] + (meters * j)
                y2 = boundaries['min_y'] + (meters * (j + 1))
                polygon_desc = {}

                # Create shape (polygon)
                p = Polygon([(x1, y1), (x1, y2), (x2, y2), (x2, y1)])

                # s = boros_shape.intersection(p)
                s = shape.intersects(p)

                # if(s.area>0):
                if s:

                    # shape.intersection(p) ATTENTION! If you use the intersection than the crawler fails!
                    polygon_desc['geometry'] = p
                    polygons.append(polygon_desc)

        gdf = gpd.GeoDataFrame(polygons, crs=tmp_crs)
        gdf = gdf.reset_index().rename(columns={"index": constants.TILE_ID})

        # Convert TILE_ID to have str type
        gdf[constants.TILE_ID] = gdf[constants.TILE_ID].astype('str')

        return gdf.to_crs(crs)


# Register the builder
tiler.register_tiler('squared', SquaredTessellationTiler())




import numpy as np
from tqdm import tqdm
import operator
import pandas as pd
from ..utils import gislib, constants, utils
from ..core.flowdataframe import FlowDataFrame

distfunc = gislib.getDistance


class Radiation:
    """Radiation model.
    
    The radiation model for human migration. The radiation model assumes that the choice of a traveler's destination consists of two steps. First, each opportunity in every location is assigned a fitness represented by a number :math:`z`, chosen from some distribution :math:`P(z)` whose value represents the quality of the opportunity for the traveler. Second, the traveler ranks all opportunities according to their distances from the origin location and chooses the closest opportunity with a fitness higher than the traveler's fitness threshold, which is another random number extracted from the fitness distribution :math:`P(z)`. As a result, the average number of travelers from location :math:`i` to location :math:`j` takes the form [SGMB2012]_:
    
    .. math:: 
        T_{ij} = O_i \\frac{1}{1 - \\frac{m_i}{M}}\\frac{m_i m_j}{(m_i + s_{ij})(m_i + m_j + s_{ij})}.
        
    The destination of the :math:`O_i` trips originating in :math:`i` is sampled from a distribution of probabilities that a trip originating in :math:`i` ends in location :math:`j`. This probability depends on the number of opportunities at the origin :math:`m_i`, at the destination :math:`m_j` and the number of opportunities :math:`s_{ij}` in a circle of radius :math:`r_{ij}` centered in :math:`i` (excluding the source and destination). This conditional probability needs to be normalized so that the probability that a trip originating in the region of interest ends in this region is equal to 1. In case of a finite system it is possible to show that this is equal to :math:`1 - \\frac{m_i}{M}` where :math:`M=\\sum_i m_i` is the total number of opportunities. In the original version of the radiation model, the number of opportunities is approximated by the population, but the total inflows :math:`D_j` to each destination can also be used.

    .. image:: https://media.springernature.com/full/springer-static/image/art%3A10.1038%2Fnature10856/MediaObjects/41586_2012_Article_BFnature10856_Fig1_HTML.jpg?as=webp
    *(a)* To demonstrate the limitations of the gravity law we highlight two pairs of counties, one in Utah (UT) and the other in Alabama (AL), with similar origin (:math:`m`, blue) and destination (:math:`n`, green) populations and comparable distance :math:`r` between them (see bottom left table). The US census 2000 reports a flux that is an order of magnitude greater between the Utah counties, a difference correctly captured by the radiation model *(b, c)*. *(b)* The definition of the radiation model: an individual (for example, living in Saratoga County, New York) applies for jobs in all counties and collects potential employment offers. The number of job opportunities in each county (:math:`j`) is :math:`n_j / n_{jobs}`, chosen to be proportional to the resident population :math:`n_j`. Each offer's attractiveness (benefit) is represented by a random variable with distribution :math:`P(z)`, the numbers placed in each county representing the best offer among the :math:`n_j / n_{jobs}` trials in that area. Each county is marked in green (red) if its best offer is better (lower) than the best offer in the home county (here :math:`z = 10`). *(c)* An individual accepts the closest job that offers better benefits than his home county. In the shown configuration the individual will commute to Oneida County, New York, the closest county whose benefit :math:`z = 13` exceeds the home county benefit :math:`z = 10`. This process is repeated for each potential commuter, choosing new benefit variables :math:`z` in each case. Figure from [SGMB2012]_.
    
    Parameters
    ----------
    name : str, optional
        the name of the instantiation of the radiation model. The default is 'Radiation model'. 

    Attributes
    ----------
    name : str
        the name of the instantiation of the model.
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.utils import utils, constants
    >>> import pandas as pd
    >>> import geopandas as gpd
    >>> import numpy as np
    >>> from skmob.models import Radiation
    >>> # load a spatial tessellation
    >>> url_tess = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
    >>> tessellation = gpd.read_file(url_tess).rename(columns={'tile_id': 'tile_ID'})
    >>> print(tessellation.head())
      tile_ID  population                                           geometry
    0   36019       81716  POLYGON ((-74.006668 44.886017, -74.027389 44....
    1   36101       99145  POLYGON ((-77.099754 42.274215, -77.0996569999...
    2   36107       50872  POLYGON ((-76.25014899999999 42.296676, -76.24...
    3   36059     1346176  POLYGON ((-73.707662 40.727831, -73.700272 40....
    4   36011       79693  POLYGON ((-76.279067 42.785866, -76.2753479999...    
    >>> # load real flows into a FlowDataFrame
    >>> # download the file with the real fluxes from: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_commuting_flows_2011.csv
    >>> fdf = skmob.FlowDataFrame.from_file("NY_commuting_flows_2011.csv", 
                                            tessellation=tessellation, 
                                            tile_id='tile_ID', 
                                            sep=",")
    >>> print(fdf.head())
         flow origin destination
    0  121606  36001       36001
    1       5  36001       36005
    2      29  36001       36007
    3      11  36001       36017
    4      30  36001       36019    
    >>> # compute the total outflows from each location of the tessellation (excluding self loops)
    >>> tot_outflows = fdf[fdf['origin'] != fdf['destination']].groupby(by='origin', axis=0)['flow'].sum().fillna(0).values
    >>> tessellation[skmob.constants.TOT_OUTFLOW] = tot_outflows
    >>> print(tessellation.head())
      tile_id  population                                           geometry  \
    0   36019       81716  POLYGON ((-74.006668 44.886017, -74.027389 44....   
    1   36101       99145  POLYGON ((-77.099754 42.274215, -77.0996569999...   
    2   36107       50872  POLYGON ((-76.25014899999999 42.296676, -76.24...   
    3   36059     1346176  POLYGON ((-73.707662 40.727831, -73.700272 40....   
    4   36011       79693  POLYGON ((-76.279067 42.785866, -76.2753479999...   
       tot_outflow  
    0        29981  
    1         5319  
    2       295916  
    3         8665  
    4         8871 
    >>> np.random.seed(0)
    >>> radiation = Radiation()
    >>> rad_flows = radiation.generate(tessellation, tile_id_column='tile_ID',  tot_outflows_column='tot_outflow', relevance_column='population', out_format='flows_sample')
    >>> print(rad_flows.head())
      origin destination   flow
    0  36019       36033  11648
    1  36019       36031   4232
    2  36019       36089   5598
    3  36019       36113   1596
    4  36019       36041    117
    
    References
    ----------
    .. [SGMB2012] Simini, F., Gonzàlez, M. C., Maritan, A. & Barabasi, A.-L. (2012) A universal model for mobility and migration patterns. Nature 484(7392), 96-100, https://www.nature.com/articles/nature10856

    """

    def __init__(self, name='Radiation model'):
        self.name_ = name
        self._spatial_tessellation = None
        self._out_format = None

    def _get_flows(self, origin, total_relevance):
        """
        Compute the edges (flows or probabilities) from location `origin` to all other locations.

        Parameters
        ----------
        origin  :  int or str
            identifier of the origin location

        location2info : dict
            information of the locations

        total_relevance : float
            sum of all relevances

        distance_f  :  callable
            distance function
            default: getDistanceByHaversine

        Returns
        -------
        edges : numpy array
            the edges generated from `origin` to the other locations

        Notes
        ------
        `m`  :  relevance of origin
        `n`  :  relevance of destination
        `s`  :  relevance in the circle between origin and destination

        """
        edges = []
        probs = []

        origin_lat, origin_lng = self.lats_lngs[origin]
        origin_relevance = self.relevances[origin]

        try:
            origin_outflow = self.tot_outflows[origin]
        except AttributeError:
            origin_outflow = 1

        if origin_outflow > 0.0:

            # compute the normalization factor
            normalization_factor = 1.0 / (1.0 - origin_relevance / total_relevance)

            destinations_and_distances = []
            for destination, (dest_lat, dest_lng) in enumerate(self.lats_lngs):
                if destination != origin:
                    destinations_and_distances += \
                        [(destination, distfunc((origin_lat, origin_lng), (dest_lat, dest_lng)))]

            # sort the destinations by distance (from the closest to the farthest)
            destinations_and_distances.sort(key=operator.itemgetter(1))

            sum_inside = 0.0
            for destination, _ in destinations_and_distances:
                destination_relevance = self.relevances[destination]
                prob_origin_destination = normalization_factor * \
                                          (origin_relevance * destination_relevance) / \
                                          ((origin_relevance + sum_inside) * (
                                                      origin_relevance + sum_inside + destination_relevance))

                sum_inside += destination_relevance
                edges += [[origin, destination]]
                probs.append(prob_origin_destination)

            probs = np.array(probs)

            if self._out_format == 'flows_average':
                quantities = np.rint(origin_outflow * probs)
            elif self._out_format == 'flows_sample':
                quantities = np.random.multinomial(origin_outflow, probs)
            else:
                quantities = probs

            edges = [edges[i] + [od] for i, od in enumerate(quantities)]

        return edges

    def generate(self, spatial_tessellation, tile_id_column=constants.TILE_ID,
                 tot_outflows_column=constants.TOT_OUTFLOW,
                 relevance_column=constants.RELEVANCE, out_format='flows_average'):
        """
        Start the simulation of the Radiation model.
        
        Parameters
        ----------
        spatial_tessellation : GeoDataFrame
            the spatial tessellation on which to perform the simulation. 
        
        tile_id_column : str, optional
            the column in `spatial_tessellation` of the location identifier. The default is `constants.TILE_ID`.
            
        tot_outflows_column : str, optional
            the column in `spatial_tessellation` with the outflow of the location. The default is `constants.TOT_OUTFLOW`.
            
        relevance_column : str, optional
            the column in `spatial_tessellation` with the relevance of the location. The default is `constants.RELEVANCE`.
            
        out_format : str, optional
            the format of the generated flows. Possible values are: "flows_sample" (the number of migrations generation by a single execution of the model), "flows_average" (average number of migrations between two locations) and "probs" (probability of movement between two locations). The default is "flows_average".
            
        Returns
        -------
        FlowDataFrame
            the fluxes generated by the Radiation model.
        """
        self._out_format = out_format
        self._tile_id_column = tile_id_column
        self.lats_lngs = spatial_tessellation.geometry.apply(utils.get_geom_centroid, args=[True]).values
        self.relevances = spatial_tessellation[relevance_column].fillna(0).values
        if 'flows' in out_format:
            if tot_outflows_column not in spatial_tessellation.columns:
                raise KeyError(
                    "The column %s for the 'tot_outflows' must be present in the tessellation." % tot_outflows_column)
            self.tot_outflows = spatial_tessellation[tot_outflows_column].fillna(0).values

        # check if arguments are valid
        if out_format not in ['flows_average', 'flows_sample', 'probs']:
            raise ValueError(
                'Value of out_format "%s" is not valid. \nValid values: flows_average, flows_sample, probs.' % out_format)

        # compute the total relevance, i.e., the sum of relevances of all the locations
        total_relevance = np.sum(self.relevances)

        all_flows = []
        for origin in tqdm(range(len(spatial_tessellation))):  # tqdm print a progress bar

            # get the edges for the current origin location
            flows_from_origin = self._get_flows(origin, total_relevance)

            if len(flows_from_origin) > 0:
                all_flows += list(flows_from_origin)

        # return np.array(all_flows)
        if 'flows' in out_format:
            return self._from_matrix_to_flowdf(all_flows, spatial_tessellation)
        else:
            return all_flows

    def _from_matrix_to_flowdf(self, all_flows, spatial_tessellation):
        index2tileid = dict([(i, tileid) for i, tileid in enumerate(spatial_tessellation[self._tile_id_column].values)])
        output_list = [[index2tileid[i], index2tileid[j], flow] for i, j, flow in all_flows if flow > 0.]
        return FlowDataFrame(output_list, origin=0, destination=1, flow=2,
                             tile_id=self._tile_id_column, tessellation=spatial_tessellation)



import datetime
from tqdm import tqdm
import operator
from collections import defaultdict, Counter
import pandas as pd
import numpy as np
from ..utils import constants
import random

latitude = constants.LATITUDE
longitude = constants.LONGITUDE
date_time = constants.DATETIME
user_id = constants.UID


class MarkovDiaryGenerator:
    """Markov Diary Learner and Generator.
    
    A diary generator :math:`G` produces a mobility diary, :math:`D(t)`, containing the sequence of trips made by an agent during a time period divided in time slots of :math:`t` seconds. For example, :math:`G(3600)` and :math:`G(60)` produce mobility diaries with temporal resolutions of one hour and one minute, respectively [PS2018]_. 
    
    A Mobility Diary Learner (MDL) is a data-driven algorithm to compute a mobility diary :math:`MD` from the mobility trajectories of a set of real individuals. We use a Markov model to describe the probability that an individual follows her routine and visits a typical location at the usual time, or she breaks the routine and visits another location. First, MDL translates mobility trajectory data of real individuals into abstract mobility trajectories. Second, it uses the obtained abstract trajectory data to compute the transition probabilities of the Markov model :math:`MD(t)` [PS2018]_.
    
    Parameters
    ----------
    name : str, optional
        name of the instantiation of the class. The default is "Markov diary".
    
    Attributes
    ----------
    name : str
        name of the instantiation of the class. 
        
    markov_chain_ : dict
        the trained markov chain.

    time_slot_length : str
        length of the time slot (1h).
    
    Examples
    --------
    >>> import skmob
    >>> import pandas as pd
    >>> import geopandas as gpd
    >>> from skmob.models.epr import Ditras
    >>> from skmob.models.markov_diary_generator import MarkovDiaryGenerator
    >>> from skmob.preprocessing import filtering, compression, detection, clustering
    >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
    >>> 
    >>> df = pd.read_csv(url, sep=',', compression='gzip')
    >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
    >>> 
    >>> ctdf = compression.compress(tdf)
    >>> stdf = detection.stops(ctdf)
    >>> cstdf = clustering.cluster(stdf)
    >>> 
    >>> mdg = MarkovDiaryGenerator()
    >>> mdg.fit(cstdf, 2, lid='cluster')
    >>> 
    >>> start_time = pd.to_datetime('2019/01/01 08:00:00')
    >>> diary = mdg.generate(100, start_time)
    >>> print(diary)
                 datetime  abstract_location
    0 2019-01-01 08:00:00                  0
    1 2019-01-02 19:00:00                  1
    2 2019-01-02 20:00:00                  0
    3 2019-01-03 17:00:00                  1
    4 2019-01-03 18:00:00                  2
    5 2019-01-04 08:00:00                  0
    6 2019-01-05 03:00:00                  1
    
    References
    ----------
    .. [PS2018] Pappalardo, L. & Simini, F. (2018) Data-driven generation of spatio-temporal routines in human mobility. Data Mining and Knowledge Discovery 32, 787-829, https://link.springer.com/article/10.1007/s10618-017-0548-4
    
    See Also
    --------
    Ditras
    """
    def __init__(self, name='Markov diary'):
        self._markov_chain_ = None
        self._time_slot_length = '1h'
        self._name = name

    @property
    def markov_chain_(self):
        return self._markov_chain_

    @property
    def time_slot_length(self):
        return self._time_slot_length

    @property
    def name(self):
        return self._name

    def _create_empty_markov_chain(self):
        """
        Create an empty Markov chain, i.e., a matrix 48 * 48 where an element M(i,j) is a pair of pairs
        ((h_i, b_i), (h_j, b_j)), h_i, h_j \in {0, ..., 23} and b_i, b_j \in {0, 1}
        """
        self._markov_chain_ = defaultdict(lambda: defaultdict(float))
        for h1 in range(0, 24):
            for r1 in [0, 1]:
                for h2 in range(0, 24):
                    for r2 in [0, 1]:
                        self._markov_chain_[(h1, r1)][(h2, r2)] = 0.0

    @staticmethod
    def _select_loc(individual_df, location2frequency, location_column='location'):

        if isinstance(individual_df[location_column], str): # if there is at least a location in the time slot
            locations = individual_df[location_column].split(',')

            if len(locations) == 1:  # if there just one location, then assign that location to the time slot
                return locations[0]

            elif len(set(Counter(locations).values())) == 1:  # if there are multiple locations, with the same frequency
                return sorted({k: location2frequency[k] for k in locations}.items(),
                              key=operator.itemgetter(1),
                              reverse=True)[0][0]  # return the locations with the highest overall frequency

            else: # if there multiple location in the same slot, with different frequencies
                l = sorted(Counter(locations).items(), key=operator.itemgetter(1), reverse=True)
                return l[0][0]

        # if there is no location in the time slot, return NaN
        return np.nan

    @staticmethod
    def _get_location2frequency(traj, location_column='location'):
        """
        Compute the visitation frequency and rank of each location of an individual
        
        Parameters
        ----------
        traj : pandas DataFrame
            the trajectories of the individuals.
        
        location_column : str, optional
            the name of the column containing the location identifier. The default is "location".
            
        Returns
        -------
        tuple
            a tuple of two dictionaries of locations to the corresponding visitation frequency and rank, respectively.
        """
        location2frequency, location2rank = defaultdict(int), defaultdict(int)
        for i, row in traj.iterrows():
            if isinstance(row[location_column], str):  # if it is not NaN
                for location in row[location_column].split(','):
                    # we can have more than one location in a time slot
                    # so, every slot has a comma separated list of locations
                    location2frequency[location] += 1

        # compute location2rank
        rank = 1
        for loc in sorted(location2frequency.items(), key=operator.itemgetter(1), reverse=True):
            location, frequency = loc
            location2rank[location] = rank
            rank += 1
        return location2frequency, location2rank

    def _create_time_series(self, traj, lid='location'):#start_date, end_date, lid='location'):
        """
        Parameters
        ----------
        traj : pandas DataFrame
            the trips of an individual.
        
        lid : str, optional
            the name of the column containing the location identifier. The default is "location".
        
        Returns
        -------
        pandas DataFrame
            the time series of the abstract locations visited by the individual.
        """
        # lat_lng_df = traj[['lat', 'lng']].drop_duplicates(subset=['lat', 'lng'])
        # lat_lng_df[lid] = np.arange(len(lat_lng_df)).astype('str')
        # traj = pd.merge(traj, lat_lng_df, on=['lat', 'lng'])

        shift = traj[constants.DATETIME].min().hour
        traj = traj[[constants.DATETIME, lid]].set_index(date_time)
        traj[lid] = traj[lid].astype('str')
        # enlarge (eventually) the time series with the specified freq (replace empty time slots with NaN)
        traj = traj.groupby(pd.Grouper(freq=self._time_slot_length, closed='left')).aggregate(lambda x: ','.join(x)).replace('', np.nan)


        # compute the frequency of every location visited by the individual
        location2frequency, location2rank = self._get_location2frequency(traj, location_column=lid)

        # select the location for every slot
        # ix = pd.DatetimeIndex(start=start_date, end=end_date, freq=self._time_slot_length)
        #ix = pd.date_range(start=start_date, end=end_date, freq=self._time_slot_length)
        time_series = traj.apply(lambda x: self._select_loc(x, location2frequency, location_column=lid), axis=1)#.reindex(ix)

        # fill the slots with NaN with the previous element or the next element ###
        time_series.fillna(method='ffill', inplace=True)
        time_series.fillna(method='bfill', inplace=True)

        # you can use location2frequency to assign a number to every location
        time_series = time_series.apply(lambda x: location2rank[x])
        return time_series, shift

    def _update_markov_chain(self, time_series, shift=0):
        """
        Update the Markov Chain by including the behavior of an individual
        
        Parameters
        ----------
        time_series: pandas DataFrame
            time series of abstract locations visisted by an individual.
        """
        HOME = 1
        TYPICAL, NON_TYPICAL = 1, 0

        n = len(time_series)  # n is the length of the time series of the individual
        slot = 0  # it starts from the first slot in the time series

        while slot < n - 1:  # scan the time series of the individual, time slot by time slot

            #h = (slot % 24)
            h = (slot  + shift) % 24  # h, the hour of the day
            next_h = (h + 1) % 24  # next_h, the next hour of the day

            loc_h = time_series[slot]  # loc_h  ,   abstract location at the current slot
            next_loc_h = time_series[slot + 1]  # d_{h+1},   abstract location at the next slot

            if loc_h == HOME:  # if \delta(loc_h, t_h) == 1, i.e., she stays at home

                # we have two cases
                if next_loc_h == HOME:  # if \delta(d_{h + 1}, t_{h + 1}) == 1

                    # we are in Type1: (h, 1) --> (h + 1, 1)
                    self._markov_chain_[(h, TYPICAL)][(next_h, TYPICAL)] += 1

                else:  # she will be not in the typical location

                    # we are in Type2: (h, 1) --> (h + tau, 0)
                    tau = 1
                    if slot + 2 < n:  # if slot is the second last in the time series

                        for j in range(slot + 2, n):  # in slot + 1 we do not have HOME so we start from slot + 2
                            loc_hh = time_series[j]
                            if loc_hh == next_loc_h:  # if \delta(d_{h + j}, d_{h + 1}) == 1
                                tau += 1
                            else:
                                break

                        h_tau = (h + tau) % 24
                        # update the state of edge (h, 1) --> (h + tau, 0)
                        self._markov_chain_[(h, TYPICAL)][(h_tau, NON_TYPICAL)] += 1
                        slot = j - 2 #1

                    else:  # terminate the while cycle
                        slot = n

            else:  # loc_h != HOME

                if next_loc_h == HOME:  # if \delta(d_{h + 1}, t_{h + 1}) == 1, i.e., she will stay at home

                    # we are in Type3: (h, 0) --> (h + 1, 1)
                    self._markov_chain_[(h, NON_TYPICAL)][(next_h, TYPICAL)] += 1

                else:

                    # we are in Type 4: (h, 0) --> (h + tau, 0)
                    tau = 1
                    if slot + 2 < n:

                        for j in range(slot + 2, n):
                            loc_hh = time_series[j]
                            if loc_hh == next_loc_h:  # if \delta(d_{h + j}, d_{h + 1}) == 1
                                tau += 1
                            else:
                                break

                        h_tau = (h + tau) % 24

                        # update the state of edge (h, 0) --> (h + tau, 0)
                        self._markov_chain_[(h, NON_TYPICAL)][(h_tau, NON_TYPICAL)] += 1
                        slot = j - 2 #1

                    else:
                        slot = n

            slot += 1

    def _normalize_markov_chain(self):
        """
        Transform the dictionary into a proper Markov chain, i.e., normalize by row in order
        to obtain transition probabilities.
        """
        # compute the probabilities of the Markov chain, i.e. normalize by row
        for state1 in self._markov_chain_:
            tot = sum([prob for prob in self._markov_chain_[state1].values()])
            for state2 in self._markov_chain_[state1]:
                if tot != 0.0:
                    self._markov_chain_[state1][state2] /= tot

    def fit(self, traj, n_individuals, lid='location'): #start_date, end_date
        """
        Train the markov mobility diary from real trajectories.
        
        Parameters
        ----------
        traj : TrajDataFrame
            the trajectories of the individuals.
        
        n_individuals : int
            the number of individuals in the TrajDataFrame to consider.
        
        lid : string, optional
            the name of the column containing the identifier of the location. The default is "location".
        """
        self._create_empty_markov_chain()  # initialize the markov chain

        individuals = traj.uid.unique()  # list of individuals' identifiers
        with tqdm(total=n_individuals) as pbar:

            for individual in individuals[:n_individuals]:

                # create the time series of the individual
                time_series, shift = self._create_time_series(traj[traj.uid == individual], lid=lid) # start_date, end_date,

                # update the markov chain according to the individual's time series
                self._update_markov_chain(time_series, shift)

                pbar.update(1)

        # normalize the markov chain, i.e., normalize the transitions by row
        self._normalize_markov_chain()

    @staticmethod
    def _weighted_random_selection(weights):
        """
        Choose an index from the list of weights according to the numbers in the list

        Parameters
        ----------
        weights: list
            a list of weights (e.g., probabilities)

        Returns
        -------
        index: int
            the index of element chosen from the list
        """
        # totals = []
        # running_total = 0
        #
        # for w in weights:
        #     running_total += w
        #     totals.append(running_total)
        #
        # rnd = random.random() * running_total
        # for index, total in enumerate(totals):
        #     if rnd < total:
        #         return index
        # return len(totals) - 1
        return np.searchsorted(np.cumsum(weights), random.random())


    def generate(self, diary_length, start_date):
        """
        Start the generation of the mobility diary.
        
        Parameters
        ----------
        diary_length : int
            the length of the diary in hours.

        start_date : datetime
            the starting date of the generation.
        
        Returns
        -------
        pandas DataFrame
            the generated mobility diary.
        """
        current_date = start_date
        V, i = [], 0
        prev_state = (i, 1)  # it starts from the typical location at midnight
        V.append(prev_state)

        while i < diary_length:

            h = i % 24  # the hour of the day

            # select the next state in the Markov chain
            p = list(self._markov_chain_[prev_state].values())
            if sum(p) == 0.:
                hh, rr = prev_state
                next_state = ((hh + 1) % 24, rr)
            else:
                index = self._weighted_random_selection(p)
                next_state = list(self._markov_chain_[prev_state].keys())[index]
            V.append(next_state)

            j = next_state[0]
            if j > h:  # we are in the same day
                i += j - h
            else:  # we are in the next day
                i += 24 - h + j

            prev_state = next_state

        # now we translate the temporal diary into the the mobility diary
        prev, diary, other_count = V[0], [], 1
        diary.append([current_date, 0])

        for v in V[1:]:  # scan all the states obtained and create the synthetic time series
            h, s = v
            h_prev, s_prev = prev

            if s == 1:  # if in that hour she visits home
                current_date += datetime.timedelta(hours=1)
                diary.append([current_date, 0])
                other_count = 1
            else:  # if in that hour she does NOT visit home

                if h > h_prev:  # we are in the same day
                    j = h - h_prev
                else:  # we are in the next day
                    j = 24 - h_prev + h

                for i in range(0, j):
                    current_date += datetime.timedelta(hours=1)
                    diary.append([current_date, other_count])
                other_count += 1

            prev = v

        short_diary = []
        prev_location = -1
        for visit_date, abstract_location in diary[0: diary_length]:
            if abstract_location != prev_location:
                short_diary.append([visit_date, abstract_location])
            prev_location = abstract_location

        diary_df = pd.DataFrame(short_diary, columns=[date_time, 'abstract_location'])
        return diary_df



import pandas as pd
from ..utils import constants, plot, utils
import numpy as np
from warnings import warn
from shapely.geometry import Polygon, Point
import geopandas as gpd
from .flowdataframe import FlowDataFrame
from skmob.preprocessing import routing


class TrajSeries(pd.Series):

    @property
    def _constructor(self):
        return TrajSeries

    @property
    def _constructor_expanddim(self):
        return TrajDataFrame


class TrajDataFrame(pd.DataFrame):
    """TrajDataFrame.
    
    A TrajDataFrame object is a pandas.DataFrame that has three columns latitude, longitude and datetime. TrajDataFrame accepts the following keyword arguments:
    
    Parameters
    ----------
    data : list or dict or pandas DataFrame
        the data that must be embedded into a TrajDataFrame.
        
    latitude : int or str, optional
        the position or the name of the column in `data` containing the latitude. The default is `constants.LATITUDE`.
        
    longitude : int or str, optional
        the position or the name of the column in `data` containing the longitude. The default is `constants.LONGITUDE`.
        
    datetime : int or str, optional
        the position or the name of the column in `data` containing the datetime. The default is `constants.DATETIME`.
        
    user_id : int or str, optional
        the position or the name of the column in `data`containing the user identifier. The default is `constants.UID`.
        
    trajectory_id : int or str, optional
        the position or the name of the column in `data` containing the trajectory identifier. The default is `constants.TID`.
        
    timestamp : boolean, optional
        it True, the datetime is a timestamp. The default is `False`.
        
    crs : dict, optional
        the coordinate reference system of the geographic points. The default is `{"init": "epsg:4326"}`.
        
    parameters : dict, optional
        parameters to add to the TrajDataFrame. The default is `{}` (no parameters).
        
    Examples
    --------
    >>> import skmob
    >>> # create a TrajDataFrame from a list
    >>> data_list = [[1, 39.984094, 116.319236, '2008-10-23 13:53:05'], [1, 39.984198, 116.319322, '2008-10-23 13:53:06'], [1, 39.984224, 116.319402, '2008-10-23 13:53:11'], [1, 39.984211, 116.319389, '2008-10-23 13:53:16']]
    >>> tdf = skmob.TrajDataFrame(data_list, latitude=1, longitude=2, datetime=3)
    >>> print(tdf.head())
       0        lat         lng            datetime
    0  1  39.984094  116.319236 2008-10-23 13:53:05
    1  1  39.984198  116.319322 2008-10-23 13:53:06
    2  1  39.984224  116.319402 2008-10-23 13:53:11
    3  1  39.984211  116.319389 2008-10-23 13:53:16
    >>> print(type(tdf))
    <class 'skmob.core.trajectorydataframe.TrajDataFrame'>
    >>> 
    >>> # create a TrajDataFrame from a pandas DataFrame
    >>> import pandas as pd
    >>> # create a DataFrame from the previous list 
    >>> data_df = pd.DataFrame(data_list, columns=['user', 'latitude', 'lng', 'hour'])
    >>> print(type(data_df))
    <class 'pandas.core.frame.DataFrame'>
    >>> tdf = skmob.TrajDataFrame(data_df, latitude='latitude', datetime='hour', user_id='user')
    >>> print(type(tdf))
    <class 'skmob.core.trajectorydataframe.TrajDataFrame'>
    >>> print(tdf.head())
       uid        lat         lng            datetime
    0    1  39.984094  116.319236 2008-10-23 13:53:05
    1    1  39.984198  116.319322 2008-10-23 13:53:06
    2    1  39.984224  116.319402 2008-10-23 13:53:11
    3    1  39.984211  116.319389 2008-10-23 13:53:16
    """
    _metadata = ['_parameters', '_crs'] # All the metadata that should be accessible must be also in the metadata method

    def __init__(self, data, latitude=constants.LATITUDE, longitude=constants.LONGITUDE, datetime=constants.DATETIME,
                 user_id=constants.UID, trajectory_id=constants.TID,
                 timestamp=False, crs={"init": "epsg:4326"}, parameters={}):

        original2default = {latitude: constants.LATITUDE,
                            longitude: constants.LONGITUDE,
                            datetime: constants.DATETIME,
                            user_id: constants.UID,
                            trajectory_id: constants.TID}

        columns = None

        if isinstance(data, pd.DataFrame):
            tdf = data.rename(columns=original2default)
            columns = tdf.columns

        # Dictionary
        elif isinstance(data, dict):
            tdf = pd.DataFrame.from_dict(data).rename(columns=original2default)
            columns = tdf.columns

        # List
        elif isinstance(data, list) or isinstance(data, np.ndarray):
            tdf = data
            columns = []
            num_columns = len(data[0])
            for i in range(num_columns):
                try:
                    columns += [original2default[i]]
                except KeyError:
                    columns += [i]

        elif isinstance(data, pd.core.internals.BlockManager):
            tdf = data

        else:
            raise TypeError('DataFrame constructor called with incompatible data and dtype: {e}'.format(e=type(data)))

        super(TrajDataFrame, self).__init__(tdf, columns=columns)

        # Check crs consistency
        if crs is None:
            warn("crs will be set to the default crs WGS84 (EPSG:4326).")

        if not isinstance(crs, dict):
            raise TypeError('crs must be a dict type.')

        self._crs = crs

        if not isinstance(parameters, dict):
            raise AttributeError("parameters must be a dictionary.")

        self._parameters = parameters

        if self._has_traj_columns():
            self._set_traj(timestamp=timestamp, inplace=True)

    def _has_traj_columns(self):

        if (constants.DATETIME in self) and (constants.LATITUDE in self) and (constants.LONGITUDE in self):
            return True

        return False

    def _is_trajdataframe(self):

        if ((constants.DATETIME in self) and pd.core.dtypes.common.is_datetime64_any_dtype(self[constants.DATETIME]))\
                and ((constants.LONGITUDE in self) and pd.core.dtypes.common.is_float_dtype(self[constants.LONGITUDE])) \
                and ((constants.LATITUDE in self) and pd.core.dtypes.common.is_float_dtype(self[constants.LATITUDE])):

            return True

        return False

    def _set_traj(self, timestamp=False, inplace=False):

        if not inplace:
            frame = self.copy()
        else:
            frame = self

        if timestamp:
            frame[constants.DATETIME] = pd.to_datetime(frame[constants.DATETIME], unit='s')

        if not pd.core.dtypes.common.is_datetime64_any_dtype(frame[constants.DATETIME].dtype):
            frame[constants.DATETIME] = pd.to_datetime(frame[constants.DATETIME])

        if not pd.core.dtypes.common.is_float_dtype(frame[constants.LONGITUDE].dtype):
            frame[constants.LONGITUDE] = frame[constants.LONGITUDE].astype('float')

        if not pd.core.dtypes.common.is_float_dtype(frame[constants.LATITUDE].dtype):
            frame[constants.LATITUDE] = frame[constants.LATITUDE].astype('float')

        frame.parameters = self._parameters
        frame.crs = self._crs

        if not inplace:
            return frame

    def to_flowdataframe(self, tessellation, remove_na=False, self_loops=True):
        """
        Aggregate a TrajDataFrame into a FlowDataFrame.
        
        Parameters
        ----------
        tessellation : GeoDataFrame
            the spatial tessellation to use to aggregate the points. 
            
        remove_na : boolean 
            if True, remove the points that do not have a corresponding tile in the spatial tessellation. The default is `False`.
            
        self_loop : boolean
            if True, it counts movements that start and end in the same tile. The default is `True`.
        
        Returns
        -------
        FlowDataFrame
            the FlowDataFrame obtained as an aggregation of the TrajDataFrame
        
        Examples
        --------
        >>> import skmob
        >>> from skmob.tessellation import tilers
        >>> import pandas as pd
        >>> from skmob.preprocessing import filtering
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf.head())
                 lat         lng            datetime  uid
        0  39.984094  116.319236 2008-10-23 05:53:05    1
        1  39.984198  116.319322 2008-10-23 05:53:06    1
        2  39.984224  116.319402 2008-10-23 05:53:11    1
        3  39.984211  116.319389 2008-10-23 05:53:16    1
        4  39.984217  116.319422 2008-10-23 05:53:21    1
        >>> # build a tessellation over the city
        >>> tessellation = tilers.tiler.get("squared", base_shape="Beijing, China", meters=15000)
        >>> # remove_na enable removing points that are not contained in the tessellation
        >>> fdf = tdf.to_flowdataframe(tessellation=tessellation, self_loops=True, remove_na=True)
        >>> print(fdf.head())
          origin destination  flow
        0     49          49   788
        1     49          62     1
        2     50          50  4974
        3     50          63     1
        4     61          61   207
        
        See Also
        --------
        FlowDataFrame
        """

        # Step 1: order the dataframe by user_id, traj_id, datetime
        self.sort_values(by=self.__operate_on(), ascending=True, inplace=True)

        # Step 2: map the trajectory onto the tessellation
        flow = self.mapping(tessellation, remove_na=remove_na)

        # Step 3: groupby tile_id and sum to obtain the flow
        flow.loc[:, constants.DESTINATION] = flow[constants.TILE_ID].shift(-1)
        flow = flow.groupby([constants.TILE_ID, constants.DESTINATION]).size().reset_index(name=constants.FLOW)
        flow.rename(columns={constants.TILE_ID: constants.ORIGIN}, inplace=True)

        if not self_loops:
            flow = flow[flow[constants.ORIGIN] != flow[constants.DESTINATION]]

        return FlowDataFrame(flow, tessellation=tessellation)

    def to_geodataframe(self):

        gdf = gpd.GeoDataFrame(self.copy(), geometry=gpd.points_from_xy(self[constants.LONGITUDE],
                                                                        self[constants.LATITUDE]), crs=self._crs)

        return gdf

    def mapping(self, tessellation, remove_na=False):
        """
        Assign each point of the TrajDataFrame to the corresponding tile of a spatial tessellation.
        
        Parameters
        ----------
        tessellation : GeoDataFrame
            the spatial tessellation containing a geometry column with points or polygons.
            
        remove_na : boolean, optional
            if `True`, remove points that do not have a corresponding tile in the spatial tessellation. The default is `False`.
        
        Returns
        -------
        TrajDataFrame
            a TrajDataFrame with an additional column `tile_ID` indicating the tile identifiers.
        
        Examples
        --------
        >>> import skmob
        >>> from skmob.tessellation import tilers
        >>> import pandas as pd
        >>> from skmob.preprocessing import filtering
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf.head())
                 lat         lng            datetime  uid
        0  39.984094  116.319236 2008-10-23 05:53:05    1
        1  39.984198  116.319322 2008-10-23 05:53:06    1
        2  39.984224  116.319402 2008-10-23 05:53:11    1
        3  39.984211  116.319389 2008-10-23 05:53:16    1
        4  39.984217  116.319422 2008-10-23 05:53:21    1
        >>> # build a tessellation over the city
        >>> tessellation = tilers.tiler.get("squared", base_shape="Beijing, China", meters=15000)
        >>> mtdf = tdf.mapping(tessellation)
        >>> print(mtdf.head())
                 lat         lng            datetime  uid tile_ID
        0  39.984094  116.319236 2008-10-23 05:53:05    1      63
        1  39.984198  116.319322 2008-10-23 05:53:06    1      63
        2  39.984224  116.319402 2008-10-23 05:53:11    1      63
        3  39.984211  116.319389 2008-10-23 05:53:16    1      63
        4  39.984217  116.319422 2008-10-23 05:53:21    1      63        
        """

        gdf = self.to_geodataframe()

        if all(isinstance(x, Polygon) for x in tessellation.geometry):

            if remove_na:
                how = 'inner'
            else:
                how = 'left'

            tile_ids = gpd.sjoin(gdf, tessellation, how=how, op='within')[[constants.TILE_ID]]

        elif all(isinstance(x, Point) for x in tessellation.geometry):

            tile_ids = utils.nearest(gdf, tessellation, constants.TILE_ID)

        new_data = self._constructor(self).__finalize__(self)
        new_data = new_data.merge(tile_ids, right_index=True, left_index=True)

        return new_data

    def __getitem__(self, key):
        """
        It the result contains lat, lng and datetime, return a TrajDataFrame, else a pandas DataFrame.
        """
        result = super(TrajDataFrame, self).__getitem__(key)

        if (isinstance(result, TrajDataFrame)) and result._is_trajdataframe():
            result.__class__ = TrajDataFrame
            result.crs = self._crs
            result.parameters = self._parameters

        elif isinstance(result, TrajDataFrame) and not result._is_trajdataframe():
            result.__class__ = pd.DataFrame

        return result


    def settings_from(self, trajdataframe):
        """
        Copy the attributes from another TrajDataFrame.
        
        Parameters
        ----------
        trajdataframe : TrajDataFrame 
            the TrajDataFrame from which to copy the attributes.
            
        Examples
        --------
        >>> import skmob
        >>> import pandas as pd
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf1 = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> tdf1 = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf1.parameters)
        {}
        >>> tdf2.parameters['hasProperty'] = True
        >>> print(tdf2.parameters)
        {'hasProperty': True}
        >>> tdf1.settings_from(tdf2)
        >>> print(tdf1.parameters)
        {'hasProperty': True}
        """
        for k in trajdataframe.metadata:
            value = getattr(trajdataframe, k)
            setattr(self, k, value)

    @classmethod
    def from_file(cls, filename, latitude=constants.LATITUDE, longitude=constants.LONGITUDE, datetime=constants.DATETIME,
                  user_id=constants.UID, trajectory_id=constants.TID,
                  usecols=None, header='infer', timestamp=False, crs={"init": "epsg:4326"}, sep=",", parameters=None):
        df = pd.read_csv(filename, sep=sep, header=header, usecols=usecols)

        if parameters is None:
            # Init prop dictionary
            parameters = {'from_file': filename}

        return cls(df, latitude=latitude, longitude=longitude, datetime=datetime, user_id=user_id,
                   trajectory_id=trajectory_id, parameters=parameters, crs=crs, timestamp=timestamp)

    @property
    def lat(self):
        if constants.LATITUDE not in self:
            raise AttributeError("The TrajDataFrame does not contain the column '%s.'" % constants.LATITUDE)
        return self[constants.LATITUDE]

    @property
    def lng(self):
        if constants.LONGITUDE not in self:
            raise AttributeError("The TrajDataFrame does not contain the column '%s.'"%constants.LONGITUDE)
        return self[constants.LONGITUDE]

    @property
    def datetime(self):
        if constants.DATETIME not in self:
            raise AttributeError("The TrajDataFrame does not contain the column '%s.'"%constants.DATETIME)
        return self[constants.DATETIME]

    @property
    def _constructor(self):
        return TrajDataFrame

    @property
    def _constructor_sliced(self):
        return TrajSeries

    @property
    def _constructor_expanddim(self):
        return TrajDataFrame

    @property
    def metadata(self):

        md = ['crs', 'parameters']    # Add here all the metadata that are accessible from the object
        return md

    def __finalize__(self, other, method=None, **kwargs):

        """propagate metadata from other to self """
        # merge operation: using metadata of the left object
        if method == 'merge':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.left, name, None))

        # concat operation: using metadata of the first object
        elif method == 'concat':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.objs[0], name, None))
        else:
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other, name, None))

        return self

    def set_parameter(self, key, param):

        self._parameters[key] = param

    @property
    def crs(self):
        return self._crs

    @crs.setter
    def crs(self, crs):
        self._crs = crs

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):

        self._parameters = dict(parameters)

    def __operate_on(self):
        """
        Check which optional fields are present and return a list of them plus mandatory fields to which apply
        built-in pandas functions such as sort_values or groupby.
        :return: list
        """

        cols = []

        if constants.UID in self:
            cols.append(constants.UID)
        if constants.TID in self:
            cols.append(constants.TID)

        cols.append(constants.DATETIME)

        return cols

    # Sorting
    def sort_by_uid_and_datetime(self):
        if constants.UID in self.columns:
            return self.sort_values(by=[constants.UID, constants.DATETIME], ascending=[True, True])
        else:
            return self.sort_values(by=[constants.DATETIME], ascending=[True])

    # Plot methods
    def plot_trajectory(self, map_f=None, max_users=10, max_points=1000, style_function=plot.traj_style_function,
                        tiles='cartodbpositron', zoom=12, hex_color=-1, weight=2, opacity=0.75, start_end_markers=True):
        """
        Plot the trajectories on a Folium map.
        
        Parameters
        ----------
        map_f : folium.Map, optional
            a `folium.Map` object where the trajectory will be plotted. If `None`, a new map will be created. The default is `None`.

        max_users : int, optional
            maximum number of users whose trajectories should be plotted. The default is `10`.

        max_points : int, optional
            maximum number of points per individual to plot. The default is `1000`. If necessary, an individual's trajectory will be down-sampled to have at most `max_points` points.

        style_function : lambda function, optional
            function specifying the style (weight, color, opacity) of the GeoJson object. The default is `plot.traj_style_function`.

        tiles : str, optional
            folium's `tiles` parameter. The default is 'cartodbpositron'.

        zoom : int, optional
            the initial zoom on the map. The default is `12`.

        hex_color : str or int, optional
            hex color of the trajectory line. If `-1` a random color will be generated for each trajectory. The default is `-1`.

        weight : float, optional
            thickness of the trajectory line. The default is `2`.

        opacity : float, optional
            opacity (alpha level) of the trajectory line. The default is `0.75`.

        start_end_markers: boolean, optional
            if `True`, add markers on the start and end points of the trajectory. The default is `True`.
        
        Returns
        -------
        folium.Map
            a `folium.Map` object with the plotted trajectories.
        
        Examples
        --------
        >>> import skmob
        >>> import pandas as pd
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf.head())
                 lat         lng            datetime  uid
        0  39.984094  116.319236 2008-10-23 05:53:05    1
        1  39.984198  116.319322 2008-10-23 05:53:06    1
        2  39.984224  116.319402 2008-10-23 05:53:11    1
        3  39.984211  116.319389 2008-10-23 05:53:16    1
        4  39.984217  116.319422 2008-10-23 05:53:21    1
        >>> m = tdf.plot_trajectory(zoom=12, weight=3, opacity=0.9, tiles='Stamen Toner')
        >>> m
        
        .. image:: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/examples/plot_trajectory_example.png
        """
        return plot.plot_trajectory(self, map_f=map_f, max_users=max_users, max_points=max_points,
                                    style_function=style_function, tiles=tiles, zoom=zoom, hex_color=hex_color,
                                    weight=weight, opacity=opacity, start_end_markers=start_end_markers)

    def plot_stops(self, map_f=None, max_users=10, tiles='cartodbpositron', zoom=12, hex_color=-1, opacity=0.3,
                   radius=12, popup=True):
        """
        Plot the stops in the TrajDataFrame on a Folium map. This function requires a TrajDataFrame with stops or clusters, output of `preprocessing.detection.stops` or `preprocessing.clustering.cluster` functions. The column `constants.LEAVING_DATETIME` must be present.
        
        Parameters
        ----------
        map_f : folium.Map
            a `folium.Map` object where the trajectory will be plotted. If `None`, a new map will be created. The default is `None`.
            
        max_users : int, optional
            maximum number of users whose trajectories should be plotted. The default is `10`.
            
        tiles : str, optional
            folium's `tiles` parameter. The default is 'cartodbpositron'.

        zoom : int, optional
            the initial zoom on the map. The default is `12`.

        hex_color : str or int, optional
            hex color of the trajectory line. If `-1` a random color will be generated for each trajectory. The default is `-1`.

        opacity : float, optional
            opacity (alpha level) of the trajectory line. The default is `0.75`.

        radius : float, optional
            size of the markers. The defeault is `12`.

        popup : boolean, optional
            if `True`, when clicking on a marker a popup window displaying information on the stop will appear. The default is `True`.
        
        Returns
        -------
        folium.Map
            a `folium.Map` object with the plotted stops.
        
        Examples
        --------
        >>> import skmob
        >>> from skmob.preprocessing import detection
        >>> import pandas as pd
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf.head())
                 lat         lng            datetime  uid
        0  39.984094  116.319236 2008-10-23 05:53:05    1
        1  39.984198  116.319322 2008-10-23 05:53:06    1
        2  39.984224  116.319402 2008-10-23 05:53:11    1
        3  39.984211  116.319389 2008-10-23 05:53:16    1
        4  39.984217  116.319422 2008-10-23 05:53:21    1
        >>> stdf = detection.stops(tdf, stop_radius_factor=0.5, minutes_for_a_stop=20.0, spatial_radius_km=0.2, leaving_time=True)
        >>> print(stdf.head())
                 lat         lng            datetime  uid    leaving_datetime
        0  39.978030  116.327481 2008-10-23 06:01:37    1 2008-10-23 10:32:53
        1  40.013820  116.306532 2008-10-23 11:10:19    1 2008-10-23 23:45:27
        2  39.978419  116.326870 2008-10-24 00:21:52    1 2008-10-24 01:47:30
        3  39.981166  116.308475 2008-10-24 02:02:31    1 2008-10-24 02:30:29
        4  39.981431  116.309902 2008-10-24 02:30:29    1 2008-10-24 03:16:35 
        >>> map_f = tdf.plot_trajectory(max_points=1000, hex_color=-1, start_end_markers=False)
        >>> stdf.plot_stops(map_f=map_f, hex_color=-1)
        
        .. image:: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/examples/plot_stops_example.png
        """
        return plot.plot_stops(self, map_f=map_f, max_users=max_users, tiles=tiles, zoom=zoom,
                               hex_color=hex_color, opacity=opacity, radius=radius, popup=popup)

    def plot_diary(self, user, start_datetime=None, end_datetime=None, ax=None, legend=False):
        """
        Plot a mobility diary of an individual in a TrajDataFrame. It requires a TrajDataFrame with clusters, output of `preprocessing.clustering.cluster`. The column `constants.CLUSTER` must be present.

        Parameters
        ----------
        user : str or int
            user identifier whose diary should be plotted.

        start_datetime : datetime.datetime, optional
            only stops made after this date will be plotted. If `None` the datetime of the oldest stop will be selected. The default is `None`.

        end_datetime : datetime.datetime, optional
            only stops made before this date will be plotted. If `None` the datetime of the newest stop will be selected. The default is `None`.

        ax : matplotlib.axes, optional
            axes where the diary will be plotted. If `None` a new ax is created. The default is `None`.

        legend : bool, optional
            If `True`, legend with cluster IDs is shown. The default is `False`.
        
        Returns
        -------
        matplotlib.axes
            the `matplotlib.axes` object of the plotted diary.
        
        Examples
        --------
        >>> import skmob
        >>> from skmob.preprocessing import detection, clustering
        >>> import pandas as pd
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf.head())
                 lat         lng            datetime  uid
        0  39.984094  116.319236 2008-10-23 05:53:05    1
        1  39.984198  116.319322 2008-10-23 05:53:06    1
        2  39.984224  116.319402 2008-10-23 05:53:11    1
        3  39.984211  116.319389 2008-10-23 05:53:16    1
        4  39.984217  116.319422 2008-10-23 05:53:21    1
        >>> # detect stops
        >>> stdf = detection.stops(tdf, stop_radius_factor=0.5, minutes_for_a_stop=20.0, spatial_radius_km=0.2, leaving_time=True)
        >>> print(stdf.head())
                 lat         lng            datetime  uid    leaving_datetime
        0  39.978030  116.327481 2008-10-23 06:01:37    1 2008-10-23 10:32:53
        1  40.013820  116.306532 2008-10-23 11:10:19    1 2008-10-23 23:45:27
        2  39.978419  116.326870 2008-10-24 00:21:52    1 2008-10-24 01:47:30
        3  39.981166  116.308475 2008-10-24 02:02:31    1 2008-10-24 02:30:29
        4  39.981431  116.309902 2008-10-24 02:30:29    1 2008-10-24 03:16:35 
        >>> # cluster stops
        >>> cstdf = clustering.cluster(stdf, cluster_radius_km=0.1, min_samples=1)
        >>> print(cstdf.head())
                 lat         lng            datetime  uid    leaving_datetime  cluster
        0  39.978030  116.327481 2008-10-23 06:01:37    1 2008-10-23 10:32:53        0
        1  40.013820  116.306532 2008-10-23 11:10:19    1 2008-10-23 23:45:27        1
        2  39.978419  116.326870 2008-10-24 00:21:52    1 2008-10-24 01:47:30        0
        3  39.981166  116.308475 2008-10-24 02:02:31    1 2008-10-24 02:30:29       42
        4  39.981431  116.309902 2008-10-24 02:30:29    1 2008-10-24 03:16:35       41
        >>> # plot the diary of one individual
        >>> user = 1
        >>> start_datetime = pd.to_datetime('2008-10-23 030000')
        >>> end_datetime = pd.to_datetime('2008-10-30 030000')
        >>> ax = cstdf.plot_diary(user, start_datetime=start_datetime, end_datetime=end_datetime)
        
        .. image:: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/examples/plot_diary_example.png
        """
        return plot.plot_diary(self, user, start_datetime=start_datetime, end_datetime=end_datetime, ax=ax, legend=legend)

    def route(self, G=None, index_origin=0, index_destin=-1):
        return routing.route(self, G=G, index_origin=index_origin, index_destin=index_destin)

    def timezone_conversion(self, from_timezone, to_timezone):
        """
        Convert the timezone of the datetime in the TrajDataFrame.
        
        Parameters
        ----------
        from_timezone : str
            the current timezone of the TrajDataFrame (e.g., 'GMT').

        to_timezone : str
            the new timezone of the TrajDataFrame (e.g., 'Asia/Shanghai').
            
        Examples
        --------
        >>> import skmob
        >>> import pandas as pd
        >>> # read the trajectory data (GeoLife, Beijing, China)
        >>> url = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/geolife_sample.txt.gz'
        >>> df = pd.read_csv(url, sep=',', compression='gzip')
        >>> tdf = skmob.TrajDataFrame(df, latitude='lat', longitude='lon', user_id='user', datetime='datetime')
        >>> print(tdf.head())
                 lat         lng            datetime  uid
        0  39.984094  116.319236 2008-10-23 05:53:05    1
        1  39.984198  116.319322 2008-10-23 05:53:06    1
        2  39.984224  116.319402 2008-10-23 05:53:11    1
        3  39.984211  116.319389 2008-10-23 05:53:16    1
        4  39.984217  116.319422 2008-10-23 05:53:21    1
        >>> tdf.timezone_conversion('GMT', 'Asia/Shanghai')
        >>> print(tdf.head())
                 lat         lng  uid            datetime
        0  39.984094  116.319236    1 2008-10-23 13:53:05
        1  39.984198  116.319322    1 2008-10-23 13:53:06
        2  39.984224  116.319402    1 2008-10-23 13:53:11
        3  39.984211  116.319389    1 2008-10-23 13:53:16
        4  39.984217  116.319422    1 2008-10-23 13:53:21
        """
        self.rename(columns={'datetime': 'original_datetime'}, inplace=True)
        self['datetime'] = self['original_datetime']. \
            dt.tz_localize(from_timezone). \
            dt.tz_convert(to_timezone). \
            dt.tz_localize(None)
        self.drop(columns=['original_datetime'], inplace=True)


def nparray_to_trajdataframe(trajectory_array, columns, parameters={}):
    df = pd.DataFrame(trajectory_array, columns=columns)
    tdf = TrajDataFrame(df, parameters=parameters)
    return tdf

def _dataframe_set_geometry(self, col, timestampe=False, drop=False, inplace=False, crs=None):
    if inplace:
        raise ValueError("Can't do inplace setting when converting from"
                         " DataFrame to GeoDataFrame")
    gf = TrajDataFrame(self)

    # this will copy so that BlockManager gets copied
    return gf._set_traj() #.set_geometry(col, drop=drop, inplace=False, crs=crs)

pd.DataFrame._set_traj = _dataframe_set_geometry



import numpy as np
#try:
#    import statsmodels.api as sm
#except ImportError:
#    import statsmodels as sm
import statsmodels as sm
from statsmodels.genmod.generalized_linear_model import GLM
from tqdm import tqdm
from ..utils import gislib, constants, utils
from ..core.flowdataframe import FlowDataFrame

distfunc = gislib.getDistance


def ci(i, number_locs):
    """
    The multinomial fit is a poisson fit
    with `number_locs` more parameters needed to
    implement the normalization constraints:
    i.e. the poisson variables in each of the
    `number_locs` origin location must sum to one.

    This function selects the appropriate
    parameter relative to the normalization
    constant in each origin location.
    """
    c = list(np.zeros(number_locs))
    c[i] = 1.
    return c


def exponential_deterrence_func(x, R):
    """
    Exponential deterrence function
    """
    return np.exp(- x * R)


def powerlaw_deterrence_func(x, exponent):
    """
    Power law deterrence function
    """
    return np.power(x, exponent)


def compute_distance_matrix(spatial_tessellation, origins):
    """
    Compute the matrix of distances between origin locations and all other locations.

    Parameters
    ----------
    spatial_tessellation : GeoDataFrame
        the spatial tessellation.
    
    origins : list
        indexes of the locations of origin.
    
    Returns
    -------
    distance_matrix : numpy array with distances between locations in `origins`
        and all locations in `spatial_tessellation`
    """
    coordinates = spatial_tessellation.geometry.apply(utils.get_geom_centroid, args=[True]).values

    n = len(spatial_tessellation)
    distance_matrix = np.zeros((n, n))

    for id_i in tqdm(origins):
        lat_i, lng_i = coordinates[id_i]
        for id_j in range(id_i + 1, n):
            lat_j, lng_j = coordinates[id_j]
            distance = distfunc((lat_i, lng_i), (lat_j, lng_j))
            distance_matrix[id_i, id_j] = distance
            distance_matrix[id_j, id_i] = distance
    return distance_matrix


class Gravity:
    """Gravity model.
    
    The Gravity model of human migration. In its original formulation, the probability :math:`T_{ij}` of moving from a location :math:`i` to location :math:`j` is defined as [Z1946]_:
    
    .. math::
        T_{ij} \propto \\frac{P_i P_j}{r_{ij}} 

    where :math:`P_i` and :math:`P_j` are the population of location :math:`i` and :math:`j` and :math:`r_{ij}` is the distance between :math:`i` and :math:`j`. The basic assumptions of this model are that the number of trips leaving location :math:`i` is proportional to its population :math:`P_i`, the attractivity of location :math:`j` is also proportional to :math:`P_j`, and finally, that there is a cost effect in terms of distance traveled. These ideas can be generalized assuming a relation of the type [BBGJLLMRST2018]_:
    
    .. math::
        T_{ij} = K m_i m_j f(r_{ij})
        
    where :math:`K` is a constant, the masses :math:`m_i` and :math:`m_j` relate to the number of trips leaving location :math:`i` or the ones attracted by location :math:`j`, and :math:`f(r_{ij})`, called *deterrence function*, is a decreasing function of distance. The deterrence function :math:`f(r_{ij})` is commonly modeled with a powerlaw or an exponential form.

    **Constrained gravity models**. Some of the limitations of the gravity model can be resolved via constrained versions. For example, one may hold the number of people originating from a location :math:`i` to be a known quantity :math:`O_i`, and the gravity model is then used to estimate the destination, constituting a so-called *singly constrained* gravity model of the form:

    .. math::
        T_{ij} = K_i O_i m_j f(r_{ij}) = O_i \\frac{m_i f(r_{ij})}{\sum_k m_k f(r_{ik})}.
        
    In this formulation, the proportionality constants :math:`K_i` depend on the location of the origin and its distance to the other places considered. We can fix also the total number of travelers arriving at a destination :math:`j` as :math:`D_j = \sum_i T_{ij}`, leading to a *doubly-constrained* gravity model. For each origin-destination pair, the flow is calculated as:
    
    .. math::
        T_{ij} = K_i O_i L_j D_j f(r_{ij})
        
    where there are now two flavors of proportionality constants
    
    .. math::
        K_i = \\frac{1}{\sum_j L_j D_j f(r_{ij})}, L_j = \\frac{1}{\sum_i K_i O_i f(r_{ij})}.

    Parameters
    ----------
    deterrence_func_type : str, optional
        the deterrence function to use. The possible deterrence function are "power_law" and "exponential". The default is "power_law".

    deterrence_func_args : list, optional
        the arguments of the deterrence function. The default is [-2.0].

    origin_exp : float, optional 
        the exponent of the origin's relevance (only relevant to globally-constrained model). The default is `1.0`.

    destination_exp : float, optional 
        the explonent of the destination's relevance. The default is 1.0.

    gravity_type : str, optional
        the type of gravity model. Possible values are "singly constrained" and "globally constrained". The default is "singly constrained".
        
    name : str, optional
        the name of the instantiation of the Gravity model. The default is "Gravity model".
    
    Attributes
    ----------
    deterrence_func_type : str
        the deterrence function to use. The possible deterrence function are "power_law" and "exponential". 

    deterrence_func_args : list
        the arguments of the deterrence function. 

    origin_exp : float
        the exponent of the origin's relevance (only relevant to globally-constrained model). 

    destination_exp : float
        the explonent of the destination's relevance.

    gravity_type : str
        the type of gravity model. Possible values are "singly constrained" and "globally constrained". 
        
    name : str
        the name of the instantiation of the Gravity model. 
    
    Examples
    --------
    >>> import skmob
    >>> from skmob.utils import utils, constants
    >>> import pandas as pd
    >>> import geopandas as gpd
    >>> import numpy as np
    >>> from skmob.models import Gravity
    >>> # load a spatial tessellation
    >>> url_tess = 'https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_counties_2011.geojson'
    >>> tessellation = gpd.read_file(url_tess).rename(columns={'tile_id': 'tile_ID'})
    >>> print(tessellation.head())
      tile_ID  population                                           geometry
    0   36019       81716  POLYGON ((-74.006668 44.886017, -74.027389 44....
    1   36101       99145  POLYGON ((-77.099754 42.274215, -77.0996569999...
    2   36107       50872  POLYGON ((-76.25014899999999 42.296676, -76.24...
    3   36059     1346176  POLYGON ((-73.707662 40.727831, -73.700272 40....
    4   36011       79693  POLYGON ((-76.279067 42.785866, -76.2753479999...    
    >>> # load real flows into a FlowDataFrame
    >>> # download the file with the real fluxes from: https://raw.githubusercontent.com/scikit-mobility/scikit-mobility/master/tutorial/data/NY_commuting_flows_2011.csv
    >>> fdf = skmob.FlowDataFrame.from_file("NY_commuting_flows_2011.csv", 
                                            tessellation=tessellation, 
                                            tile_id='tile_ID', 
                                            sep=",")
    >>> print(fdf.head())
         flow origin destination
    0  121606  36001       36001
    1       5  36001       36005
    2      29  36001       36007
    3      11  36001       36017
    4      30  36001       36019    
    >>> # compute the total outflows from each location of the tessellation (excluding self loops)
    >>> tot_outflows = fdf[fdf['origin'] != fdf['destination']].groupby(by='origin', axis=0)['flow'].sum().fillna(0).values
    >>> tessellation[constants.TOT_OUTFLOW] = tot_outflows
    >>> print(tessellation.head())
      tile_id  population                                           geometry  \
    0   36019       81716  POLYGON ((-74.006668 44.886017, -74.027389 44....   
    1   36101       99145  POLYGON ((-77.099754 42.274215, -77.0996569999...   
    2   36107       50872  POLYGON ((-76.25014899999999 42.296676, -76.24...   
    3   36059     1346176  POLYGON ((-73.707662 40.727831, -73.700272 40....   
    4   36011       79693  POLYGON ((-76.279067 42.785866, -76.2753479999...   
       tot_outflow  
    0        29981  
    1         5319  
    2       295916  
    3         8665  
    4         8871 
    >>> # instantiate a singly constrained Gravity model
    >>> gravity_singly = Gravity(gravity_type='singly constrained')
    >>> print(gravity_singly)
    Gravity(name="Gravity model", deterrence_func_type="power_law", deterrence_func_args=[-2.0], origin_exp=1.0, destination_exp=1.0, gravity_type="singly constrained")
    >>> np.random.seed(0)
    >>> synth_fdf = gravity_singly.generate(tessellation, 
                                       tile_id_column='tile_ID', 
                                       tot_outflows_column='tot_outflow', 
                                       relevance_column= 'population',
                                       out_format='flows')
    >>> print(synth_fdf.head())
      origin destination  flow
    0  36019       36101   101
    1  36019       36107    66
    2  36019       36059  1041
    3  36019       36011   151
    4  36019       36123    33
    >>> # fit the parameters of the Gravity model from real fluxes
    >>> gravity_singly_fitted = Gravity(gravity_type='singly constrained')
    >>> print(gravity_singly_fitted)
    Gravity(name="Gravity model", deterrence_func_type="power_law", deterrence_func_args=[-2.0], origin_exp=1.0, destination_exp=1.0, gravity_type="singly constrained")
    >>> gravity_singly_fitted.fit(fdf, relevance_column='population')
    >>> print(gravity_singly_fitted)
    Gravity(name="Gravity model", deterrence_func_type="power_law", deterrence_func_args=[-1.9947152031914186], origin_exp=1.0, destination_exp=0.6471759552223144, gravity_type="singly constrained") 
    >>> np.random.seed(0)
    >>> synth_fdf_fitted = gravity_singly_fitted.generate(tessellation, 
                                                            tile_id_column='tile_ID', 
                                                            tot_outflows_column='tot_outflow', 
                                                            relevance_column= 'population', 
                                                            out_format='flows')
    >>> print(synth_fdf_fitted.head())
      origin destination  flow
    0  36019       36101   102
    1  36019       36107    66
    2  36019       36059  1044
    3  36019       36011   152
    4  36019       36123    33
    
    References
    ----------
    .. [Z1946] Zipf, G. K. (1946) The P 1 P 2/D hypothesis: on the intercity movement of persons. American sociological review 11(6), 677-686, https://www.jstor.org/stable/2087063?seq=1#metadata_info_tab_contents
    .. [W1971] Wilson, A. G. (1971) A family of spatial interaction models, and associated developments. Environment and Planning A 3(1), 1-32, https://econpapers.repec.org/article/pioenvira/v_3a3_3ay_3a1971_3ai_3a1_3ap_3a1-32.htm
    .. [BBGJLLMRST2018] Barbosa, H., Barthelemy, M., Ghoshal, G., James, C. R., Lenormand, M., Louail, T., Menezes, R., Ramasco, J. J. , Simini, F. & Tomasini, M. (2018) Human mobility: Models and applications. Physics Reports 734, 1-74, https://www.sciencedirect.com/science/article/abs/pii/S037015731830022X
    
    See Also
    --------
    Radiation
    """

    def __init__(self, deterrence_func_type='power_law', deterrence_func_args=[-2.0],
                 origin_exp=1.0, destination_exp=1.0, gravity_type='singly constrained',
                 name='Gravity model'):

        self._name = name
        self._deterrence_func_type = deterrence_func_type
        self._deterrence_func_args = deterrence_func_args
        self._origin_exp = origin_exp
        self._destination_exp = destination_exp
        self._gravity_type = gravity_type

        # set the deterrence function
        if self._deterrence_func_type == 'power_law':  # power law deterrence function
            self._deterrence_func = powerlaw_deterrence_func
        elif self._deterrence_func_type == 'exponential':  # exponential deterrence function
            self._deterrence_func = exponential_deterrence_func
        else:
            print('Deterrence function type "%s" not available. Power law will be used.\n'
                  'Available deterrence functions are [power_law, exponential]' % self._deterrence_func_type)
            self._deterrence_func = powerlaw_deterrence_func

    @property
    def name(self):
        return self._name

    @property
    def deterrence_func_type(self):
        return self._deterrence_func_type

    @property
    def deterrence_func_args(self):
        return self._deterrence_func_args

    @property
    def origin_exp(self):
        return self._origin_exp

    @property
    def destination_exp(self):
        return self._destination_exp

    @property
    def gravity_type(self):
        return self._gravity_type

    def __str__(self):
        return 'Gravity(name=\"%s\", deterrence_func_type=\"%s\", ' \
               'deterrence_func_args=%s, origin_exp=%s, destination_exp=%s, gravity_type=\"%s\")' % \
               (self._name, self._deterrence_func_type, self._deterrence_func_args, self._origin_exp,
                self._destination_exp, self._gravity_type)

    def compute_gravity_score(self, distance_matrix, relevances_orig, relevances_dest):
        trip_probs_matrix = self._deterrence_func(distance_matrix, *self._deterrence_func_args)
        # trip_probs_matrix = np.transpose(
        #     trip_probs_matrix * relevances ** self.destination_exp) * relevances ** self._origin_exp
        trip_probs_matrix = trip_probs_matrix * relevances_dest ** self.destination_exp * \
                            np.expand_dims(relevances_orig ** self._origin_exp, axis=1)
        # put the NaN and Inf to 0.0
        np.putmask(trip_probs_matrix, np.isnan(trip_probs_matrix), 0.0)
        np.putmask(trip_probs_matrix, np.isinf(trip_probs_matrix), 0.0)
        return trip_probs_matrix

    def generate(self, spatial_tessellation, tile_id_column=constants.TILE_ID,
                 tot_outflows_column=constants.TOT_OUTFLOW, relevance_column=constants.RELEVANCE, out_format='flows'):
        """
        Start the simulation of the Gravity model.
        
        Parameters
        ----------
        spatial_tessellation : GeoDataFrame
            the spatial tessellation on which to run the simulation.
            
        tile_id_column : str, optional
            the column in `spatial_tessellation` of the location identifier. The default is `constants.TILE_ID`.
            
        tot_outflows_column : str, optional
            the column in `spatial_tessellation` with the outflow of the location. The default is `constants.TOT_OUTFLOW`.
            
        relevance_column : str, optional
            the column in `spatial_tessellation` with the relevance of the location. The default is `constants.RELEVANCE`.
            
        out_format : str, optional
            the format of the generated flows. Possible values are "flows" and "probabilities". The default is "flows".
            
        Returns
        -------
        FlowDataFrame
            the flows generated by the Gravity model.
        """
        n_locs = len(spatial_tessellation)
        relevances = spatial_tessellation[relevance_column].fillna(0).values
        self._tile_id_column = tile_id_column
        # self._spatial_tessellation = spatial_tessellation

        if out_format not in ['flows', 'probabilities']:
            print('Output format \"%s\" not available. Flows will be used.\n'
                  'Available output formats are [flows, probabilities]' % out_format)
            out_format = "flows"

        if out_format == 'flows':
            if tot_outflows_column not in spatial_tessellation.columns:
                raise KeyError("The column 'tot_outflows' must be present in the tessellation.")
            tot_outflows = spatial_tessellation[tot_outflows_column].fillna(0).values

        # the origin locations are all locations
        origins = np.arange(n_locs)

        # compute the distances between all pairs of locations
        distance_matrix = compute_distance_matrix(spatial_tessellation, origins)

        # compute scores
        trip_probs_matrix = self.compute_gravity_score(distance_matrix, relevances, relevances)

        if self._gravity_type == 'globally constrained':  # globally constrained gravity model
            trip_probs_matrix /= np.sum(trip_probs_matrix)

            if out_format == 'flows':
                # generate random fluxes according to trip probabilities
                od_matrix = np.reshape(np.random.multinomial(np.sum(tot_outflows), trip_probs_matrix.flatten()),
                                       (n_locs, n_locs))
                return self._from_matrix_to_flowdf(od_matrix, origins, spatial_tessellation)
            else:
                # return trip_probs_matrix
                return self._from_matrix_to_flowdf(trip_probs_matrix, origins, spatial_tessellation)

        else:  # singly constrained gravity model
            # trip_probs_matrix = np.transpose(trip_probs_matrix / np.sum(trip_probs_matrix, axis=0))
            trip_probs_matrix = (trip_probs_matrix.T / np.sum(trip_probs_matrix, axis=1)).T
            # put the NaN and Inf to 0.0
            np.putmask(trip_probs_matrix, np.isnan(trip_probs_matrix), 0.0)
            np.putmask(trip_probs_matrix, np.isinf(trip_probs_matrix), 0.0)

            if out_format == 'flows':
                # generate random fluxes according to trip probabilities
                od_matrix = np.array([np.random.multinomial(tot_outflows[i], trip_probs_matrix[i]) for i in origins])
                return self._from_matrix_to_flowdf(od_matrix, origins, spatial_tessellation)
            else:
                # return trip_probs_matrix
                return self._from_matrix_to_flowdf(trip_probs_matrix, origins, spatial_tessellation)

    def _from_matrix_to_flowdf(self, flow_matrix, origins, spatial_tessellation):
        index2tileid = dict([(i, tileid) for i, tileid in enumerate(spatial_tessellation[self._tile_id_column].values)])
        output_list = [[index2tileid[i], index2tileid[j], flow]
                       for i in origins for j, flow in enumerate(flow_matrix[i]) if flow > 0.]
        return FlowDataFrame(output_list, origin=0, destination=1, flow=2,
                             tile_id=self._tile_id_column, tessellation=spatial_tessellation)

    def _update_training_set(self, flow_example):

        id_origin = flow_example[constants.ORIGIN]
        id_destination = flow_example[constants.DESTINATION]
        trips = flow_example[constants.FLOW]

        if id_origin == id_destination:
            return

        try:
            coords_origin = self.lats_lngs[self.tileid2index[id_origin]]
            weight_origin = self.weights[self.tileid2index[id_origin]]
        except KeyError:
            print('Missing information for location ' + \
                  '"%s" in spatial tessellation. Skipping ...' % id_origin)
            return

        try:
            coords_destination = self.lats_lngs[self.tileid2index[id_destination]]
            weight_destination = self.weights[self.tileid2index[id_destination]]
        except KeyError:
            print('Missing information for location ' + \
                  '"%s" in spatial tessellation. Skipping ...' % id_destination)
            return

        if weight_destination <= 0:
            return

        dist = distfunc(coords_origin, coords_destination)

        if self._gravity_type == 'globally constrained':
            sc_vars = [np.log(weight_origin)]
        else:  # if singly constrained
            sc_vars = ci(self.tileid2index[id_origin], len(self.tileid2index))

        if self._deterrence_func_type == 'exponential':
            # exponential deterrence function
            self.X += [[1.] + sc_vars + [np.log(weight_destination), - dist]]
        else:
            # power law deterrence function
            self.X += [[1.] + sc_vars + [np.log(weight_destination), np.log(dist)]]

        self.y += [float(trips)]

    def fit(self, flow_df, relevance_column=constants.RELEVANCE):
        """
        Fit the parameters of the Gravity model to the flows provided in input, using a Generalized Linear Model (GLM) with a Poisson regression [FM1982]_.

        Parameters
        ----------
        flow_df  :  FlowDataFrame 
            the real flows on the spatial tessellation. 
            
        relevance_column : str, optional
            the column in the spatial tessellation with the relevance of the location. The default is constants.RELEVANCE.

        References
        ----------
        .. [FM1982] Flowerdew, R. & Murray, A. (1982) A method of fitting the gravity model based on the Poisson distribution. Journal of regional science 22(2), 191-202, https://onlinelibrary.wiley.com/doi/abs/10.1111/j.1467-9787.1982.tb00744.x

        """
        self.lats_lngs = flow_df.tessellation.geometry.apply(utils.get_geom_centroid, args=[True]).values
        self.weights = flow_df.tessellation[relevance_column].fillna(0).values
        self.tileid2index = dict(
            [(tileid, i) for i, tileid in enumerate(flow_df.tessellation[constants.TILE_ID].values)])

        self.X, self.y = [], []  # independent (X) and dependent (y) variables

        # flow_df.progress_apply(lambda flow_example: self._update_training_set(flow_example),
        #                        axis=1)
        flow_df.apply(lambda flow_example: self._update_training_set(flow_example), axis=1)

        # Perform GLM fit
        poisson_model = GLM(self.y, self.X, family=sm.genmod.families.family.Poisson(link=sm.genmod.families.links.log))
        poisson_results = poisson_model.fit()

        # Set best fit parameters
        if self._gravity_type == 'globally constrained':
            self._origin_exp = poisson_results.params[1]
            self._destination_exp = poisson_results.params[2]
            self._deterrence_func_args = [poisson_results.params[3]]
        else:  # if singly constrained
            self._origin_exp = 1.
            self._destination_exp = poisson_results.params[-2]
            self._deterrence_func_args = [poisson_results.params[-1]]

        # we delete the instance variables we do not need anymore
        del self.X
        del self.y
